from youtube_autonomous.segments.builder.youtube.enums import YoutubeChannelId
from youtubeenhanced.objects.youtube_video import YoutubeVideo
from youtubeenhanced.enums import Quality, Language
from youtubeenhanced.utils.youtube_api import create_youtube_service, is_youtube_token_valid, start_youtube_auth_flow
from yta_general_utils.checker.url import is_valid_youtube_video
from yta_general_utils.temp import create_temp_filename
from random import randint

import requests


class YoutubeDownloader:
    """
    Singleton class.

    This object simplifies the access to our Youtube provider channels.
    It uses the Google Youtube Data V3 API and, if not available, uses
    a community API to obtain the videos.

    This object is used to download memes, stock videos, sounds, music,
    etc. from our specific Youtube channels.
    """
    __instance = None

    def __new__(cls, ignore_repeated = True):
        if not YoutubeDownloader.__instance:
            YoutubeDownloader.__instance = object.__new__(cls)
        
        return YoutubeDownloader.__instance

    def __init__(self, ignore_repeated = True):
        # TODO: Maybe implement the other 'Singleton' pattern method, I don't
        # know if this one is actually working
        if not hasattr(self, 'ignore_ids'):
            if not is_youtube_token_valid():
                print('Youtube token is not valid. Please, login to get a new valid token.')
                start_youtube_auth_flow()

            self.service = create_youtube_service()
            # This is to avoid using the same video again in the whole video. This can be
            # modified to avoid repeated videos only in each segment
            self.ignore_ids = []
            self.ignore_repeated = ignore_repeated

    def activate_ignore_repeated(self):
        """
        Sets the 'ignore_repeated' flag as True. This flag checks that a video
        that has been returned once is not returned again. This method will
        activate it, so any video can be returned only once.
        """
        self.ignore_repeated = True

    def deactivate_ignore_repeated(self):
        """
        Sets the 'ignore_repeated' flag as False. This flag checks that a video
        that has been returned once is not returned again. This method will
        deactivate it, so any video can be returned unlimited times if found.
        """
        self.ignore_repeated = False

    def add_ignored_id(self, ignore_id):
        # TODO: Maybe do any more check (?)
        if ignore_id not in self.ignore_ids:
            self.ignore_ids.append(ignore_id)

    def __search(self, keywords: str, max_results: int = 25, channel_id: str = None):
        """
        Searchs videos in the provided 'channel_id' channel, using the 'keywords'
        keywords, and returning 'max_results' results as maximum.
        """
        if not keywords:
            raise Exception('No "keywords" provided.')
        
        if not max_results:
            max_results = 25

        if not channel_id:
            raise Exception('No "channel_id" provided.')

        try:
            response_videos_list = self.service.search().list(
                part = 'snippet',
                channelId = channel_id,
                maxResults = max_results,
                order = 'relevance',  # This is the most interesting by far, using the youtube search engine
                type = 'video',
                q = keywords
            ).execute()
        except Exception as e:
            print(e)
            # We try the collaborative known alternative that should work
            no_key_url = 'https://yt.lemnoslife.com/noKey/search?part=snippet&channelId=' + channel_id + '&maxResults=' + str(max_results) + '&order=relevance&type=video&q=' + keywords + '&alt=json'
            try:
                response_videos_list = requests.get(no_key_url).json()
            except Exception as e:
                print(e)
                return []

        if response_videos_list['pageInfo']['totalResults'] == 0:
            return []
        
        return response_videos_list['items']
    
    def get_video(self, url: str):
        """
        Returns the Youtube video from the 'url' provided as a YoutubeVideo
        object.
        """
        if not url:
            raise Exception('No "url" provided.')
        
        if not isinstance(url, str):
            raise Exception('The "url" parameter provided is not a str.')
        
        return YoutubeVideo(url)

    # TODO: Maybe the '__get_videos' (?)
    
    def __get_video(self, keywords: str, channel_id: str, do_randomize: bool = False) -> YoutubeVideo:
        """
        Looks for videos in the provided 'channel_id' channel according to the also
        provided 'keywords'. This method will return a video, if found, or None if 
        not. The 'do_randomize' parameter will return a random video if True and
        more than one video is found.

        This method will ignore any repeated video if 'ignore_repeated' is True in 
        this class instance.
        """
        if not keywords:
            raise Exception('No "keywords" provided.')

        if not channel_id:
            raise Exception('No "channel_id" provided.')

        youtube_videos = self.__search(keywords, 25, channel_id)

        # We look for repeated videos to eliminate from the ones found
        if self.ignore_repeated and len(self.ignore_ids) > 0:
            index = 0
            while index < len(youtube_videos):
                video = youtube_videos[index]
                if video['id']['videoId'] in self.ignore_ids:
                    del youtube_videos[index]
                    index -= 1
                index += 1

        if len(youtube_videos) == 0:
            return None

        url = 'https://www.youtube.com/watch?v=' + youtube_videos[0]['id']['videoId']
        if do_randomize:
            url = 'https://www.youtube.com/watch?v=' + youtube_videos[randint(0, len(youtube_videos) - 1)]['id']['videoId']

        return YoutubeVideo(url)

    def __download_video(self, keywords: str, channel_id: str, include_audio = False, do_randomize: bool = False, output_filename = 'output.mp4'):
        """
        Looks for the videos in the channel with provided 'channel_id' id and that match
        the also provided 'keywords'. This method will download, if any video is found, 
        the first one (or a random one if 'do_randomize' is True).

        This method will return None if no video found nor downloaded.

        This method forces the download to FULL HD quality (1920x1080).

        This method returns the locally stored 'output_filename' (that could change due to video
        extension), or None if something went wrong.
        """
        if not keywords:
            raise Exception('No "keywords" provided.')

        if not channel_id:
            raise Exception('No "channel_id" provided.')
        
        youtube_video = self.__get_video(keywords, channel_id, do_randomize)

        if not youtube_video:
            return None

        if include_audio:
            output_filename = youtube_video.download_with_audio(Quality.FULL_HD, Language.DEFAULT, output_filename)
        else:
            output_filename = youtube_video.download(Quality.FULL_HD, output_filename)

        if self.ignore_repeated:
            self.add_ignored_id(youtube_video.id)

        return output_filename
    
    def download_this_video(self, youtube_video: YoutubeVideo, output_filename: str = None, do_include_sound: bool = True, language: Language = Language.DEFAULT):
        """
        Downloads the provided 'youtube_video' and stores it locally as 
        'output_filename' if this parameter is provided, or in a temporary file
        (that is returned by this method) if not. The 'do_include_sound' parameter
        allows the user to include the audio or not. If 'do_include_sound' is True,
        the 'language' parameter is needed to decide the audio language.
        """
        if not youtube_video:
            raise Exception('No "youtube_video" provided.')
        
        if not output_filename:
            output_filename = create_temp_filename('youtube_video.mp4')

        if not isinstance(output_filename, str):
            raise Exception('The "output_filename" provided is not a str.')

        if not language:
            language = Language.DEFAULT

        if not isinstance(language, Language):
            raise Exception('The "language" parameter provided is not a Language.')

        if do_include_sound:
            output_filename = youtube_video.download_with_audio(Quality.FULL_HD, language, output_filename)
        else:
            output_filename = youtube_video.download(Quality.FULL_HD, output_filename)

        return output_filename
    
    def download_video(self, url: str, output_filename: str = None, do_include_sound: bool = True, language: Language = Language.DEFAULT):
        """
        Downloads the Youtube video from the provided 'url' and stores it locally
        as 'output_filename' if this parameter is provided, or in a temporary file
        (that is returned by this method) if not. The 'do_include_sound' parameter
        allows the user to include the audio or not. If 'do_include_sound' is True,
        the 'language' parameter is needed to decide the audio language.
        """
        if not url:
            raise Exception('No "url" provided.')
        
        if not isinstance(url, str):
            raise Exception('The "url" parameter provided is not a str.')
        
        if not output_filename:
            output_filename = create_temp_filename('youtube_video.mp4')

        if not isinstance(output_filename, str):
            raise Exception('The "output_filename" provided is not a str.')

        if not language:
            language = Language.DEFAULT

        if not isinstance(language, Language):
            raise Exception('The "language" parameter provided is not a Language.')


        # TODO: Validate url (?)
        video = YoutubeVideo(url)
        
        if not video.is_available():
            raise Exception('Provided "url" is not a valid and/or available Youtube video.')
        
        return self.download_this_video(video, output_filename, do_include_sound, language)
    
    def __download_audio(self, keywords: str, channel_id: str, do_randomize: bool = False, output_filename = 'output.mp3'):
        """
        Looks for the videos in the channel with provided 'channel_id' id and that match
        the also provided 'keywords'. This method will download, if any video is found, 
        the first one (or a random one if 'do_randomize' is True).

        This method will return None if no video found nor downloaded.

        This method forces the download to DEFAULT LANGUAGE (the first one in the list).

        This method returns the locally stored 'output_filename' (that could change due to video
        extension), or None if something went wrong.
        """
        youtube_video = self.__get_video(keywords, channel_id, do_randomize)

        if not youtube_video:
            return None
        
        if self.ignore_repeated:
            self.add_ignored_id(youtube_video.id)

        return youtube_video.download_audio(Language.DEFAULT, output_filename)

    def download_audio(self, url: str, output_filename: str = None, language: Language = Language.DEFAULT):
        """
        Downloads the Youtube video audio from the provided 'url' and stores it 
        locally as 'output_filename' if this parameter is provided, or in a 
        temporary file (that is returned by this method) if not. The 'language' 
        parameter is needed to decide the audio language.
        """
        if not url:
            raise Exception('No "url" provided.')
        
        if not isinstance(url, str):
            raise Exception('The "url" parameter provided is not a str.')
        
        if not output_filename:
            output_filename = create_temp_filename('youtube_video.mp3')

        if not isinstance(output_filename, str):
            raise Exception('The "output_filename" provided is not a str.')

        if not language:
            language = Language.DEFAULT

        if not isinstance(language, Language):
            raise Exception('The "language" parameter provided is not a Language.')

        # TODO: Validate url (?)
        video = YoutubeVideo(url)
        output_filename = video.download_audio(language, output_filename)

        return output_filename

    def download_meme_video(self, keywords: str, include_audio: bool = True, do_randomize: bool = False, output_filename: str = create_temp_filename('meme.mp4')):
        """
        Downloads a meme from the specific Youtube channel configured as Meme channel. It will
        include the audio if 'include_audio' is True, and will return the first result (based 
        on Youtube search relevance) or a random one if 'do_randomize' is True. It will be 
        stored locally as 'output_filename' if found.

        This method returns the 'output_filename' of the locally stored video, or None if not
        downloaded.
        """
        return self.__download_video(keywords, YoutubeChannelId.MEMES.value, do_randomize = do_randomize, include_audio = include_audio, output_filename = output_filename)

    def download_stock_video(self, keywords: str, include_audio: bool = True, do_randomize: bool = False, output_filename: str = create_temp_filename('stock.mp4')):
        """
        Downloads a stock video from the specific Youtube channel configured as Stock videos 
        channel. It will include the audio if 'include_audio' is True, and will return the 
        first result (based on Youtube search relevance) or a random one if 'do_randomize' is
        True. It will be stored locally as 'output_filename' if found.

        This method returns the 'output_filename' of the locally stored video, or None if not
        downloaded.
        """
        return self.__download_video(keywords, YoutubeChannelId.STOCK.value, do_randomize = do_randomize, include_audio = include_audio, output_filename = output_filename)
    
    def download_sound_audio(self, keywords: str, output_filename: str = create_temp_filename('sound.mp3')):
        """
        Downloads a sound from the specific Youtube channel configured as Sounds channel. It 
        will be stored locally as 'output_filename' if found.

        This method returns the 'output_filename' of the locally stored audio, or None if not
        downloaded.
        """
        return self.__download_audio(keywords, YoutubeChannelId.SOUNDS.value, do_randomize = False, output_filename = output_filename)
    
    def download_music_audio(self, keywords: str, output_filename: str = create_temp_filename('music.mp3')):
        """
        Downloads a song from the specific Youtube channel configured as Music channel. It 
        will be stored locally as 'output_filename' if found.

        This method returns the 'output_filename' of the locally stored audio, or None if not
        downloaded.
        """
        return self.__download_audio(keywords, YoutubeChannelId.MUSIC.value, do_randomize = False, output_filename = output_filename)
    
    def get_stock_video(self, keywords: str, do_randomize: bool = False) -> YoutubeVideo:
        """
        Searches for the provided 'keywords' in the Stock Videos Youtube channel and gets,
        if available, a video matching those 'keywords'. It will return the first one or a
        random one if 'do_randomize' is True.
        """
        if not keywords:
            raise Exception('No "keywords" provided.')
        
        return self.__get_video(keywords, YoutubeChannelId.STOCK.value, do_randomize)
    

"""
    Pay attention to this yt-dlp link, it is interesting: 
    https://github.com/yt-dlp/yt-dlp?tab=readme-ov-file#download-options
"""


from youtube_autonomous.experimental.elements.element_data import ElementData
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER
from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.rules.effect_element_rules import ElementRules
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from yta_multimedia.audio.voice.transcription.stt.whisper import get_transcription_with_timestamps
from yta_general_utils.file import copy_file
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.temp import get_temp_filename
from yta_general_utils.logger import print_completed
from moviepy.editor import AudioFileClip, VideoFileClip
from typing import Union


class Element:
    _audio_clip = None
    _video_clip = None
    _full_clip = None

    @property
    def type(self):
        return self.data.type

    @property
    def index(self):
        return self._index
    
    @index.setter
    def index(self, index: int):
        self._index = index

    @property
    def status(self):
        return self.data.status
    
    @status.setter
    def status(self, status: str):
        self.data.status = status

    @property
    def audio_filename(self):
        audio_filename = self.data.audio_filename

        if audio_filename and not self.audio_clip:
            self.audio_clip = AudioFileClip(audio_filename)

        return audio_filename
    
    @audio_filename.setter
    def audio_filename(self, audio_filename: Union[str, None]):
        self.data.audio_filename = audio_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.data.audio_filename:
            self.audio_clip = AudioFileClip(self.data.audio_filename)

    @property
    def audio_clip(self):
        return self._audio_clip
    
    @audio_clip.setter
    def audio_clip(self, audio_clip):
        self._audio_clip = audio_clip

    @property
    def video_filename(self):
        video_filename = self.data.video_filename

        if video_filename and not self.video_clip:
            self._video_clip = VideoFileClip(video_filename)

        return video_filename
    
    @video_filename.setter
    def video_filename(self, video_filename: Union[str, None]):
        self.data.video_filename = video_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.data.video_filename:
            self.video_clip = VideoFileClip(self.data.video_filename)

    @property
    def video_clip(self):
        return self._video_clip
    
    @video_clip.setter
    def video_clip(self, video_clip):
        self._video_clip = video_clip

    @property
    def full_filename(self):
        full_filename = self.data.full_filename

        if full_filename and not self.full_clip:
            self.full_clip = VideoFileClip(full_filename)

        return full_filename
    
    @full_filename.setter
    def full_filename(self, full_filename: Union[str, None]):
        self.data.full_filename = full_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.data.full_filename:
            self.full_clip = VideoFileClip(self.data.full_filename)

    @property
    def full_clip(self):
        return self._full_clip
    
    @full_clip.setter
    def full_clip(self, full_clip):
        self._full_clip = full_clip

    @property
    def audio_narration_filename(self):
        return self.data.audio_narration_filename
    
    @audio_narration_filename.setter
    def audio_narration_filename(self, audio_narration_filename: Union[str, None]):
        self.data.audio_narration_filename = audio_narration_filename

    @property
    def transcription(self):
        return self.data.transcription
    
    @transcription.setter
    def transcription(self, transcription: Union[dict, None]):
        self.data.transcription = transcription

    @property
    def shortcodes(self):
        return self.data.shortcodes
    
    @shortcodes.setter
    def shortcodes(self, shortcodes: Union[list['Shortcode'], None]):
        self.data.shortcodes = shortcodes

    @property
    def calculated_duration(self):
        return self.data.calculated_duration
    
    @calculated_duration.setter
    def calculated_duration(self, calculated_duration: Union[float, int, None]):
        self.data.calculated_duration = calculated_duration

    @property
    def text(self):
        return self.data.text
    
    @property
    def duration(self):
        return self.data.duration

    @duration.setter
    def duration(self, duration):
        self.data.duration = duration

    @property
    def filename(self):
        return self.data.filename
    
    @property
    def url(self):
        return self.data.url
    
    @property
    def voice(self):
        return self.data.voice
    
    @property
    def narration_text(self):
        return self.data.narration_text
    
    @property
    def narration_text_sanitized_without_shortcodes(self):
        return self.data.narration_text_sanitized_without_shortcodes
    
    @narration_text_sanitized_without_shortcodes.setter
    def narration_text_sanitized_without_shortcodes(self, narration_text_sanitized_without_shortcodes: str):
        self.data.narration_text_sanitized_without_shortcodes = narration_text_sanitized_without_shortcodes

    @property
    def narration_text_with_simplified_shortcodes(self):
        return self.data.narration_text_with_simplified_shortcodes
    
    @narration_text_with_simplified_shortcodes.setter
    def narration_text_with_simplified_shortcodes(self, narration_text_with_simplified_shortcodes: str):
        self.data.narration_text_with_simplified_shortcodes = narration_text_with_simplified_shortcodes

    @property
    def narration_text_sanitized(self):
        return self.data.narration_text_sanitized
    
    @narration_text_sanitized.setter
    def narration_text_sanitized(self, narration_text_sanitized: str):
        self.data.narration_text_sanitized = narration_text_sanitized
    
    def __init__(self, index: int, data: dict):
        self.index = index
        self.data = ElementData(data)
        self.rules = ElementRules.get_subclass_by_type(data['type'])()
        self.builder = ElementBuilder.get_subclass_by_type(data['type'])()
        self.rules_checker = RulesChecker(self.rules)

        self.database_handler = DatabaseHandler()
        # TODO: Set shortcode tags please (read Notion)
        self.shortcode_parser = ShortcodeParser([])

    def create_segment_file(self, filename: str):
        """
        Creates a filename within the definitive segments folder
        to keep the generated file locally to recover it in the
        next project execution if something goes wrong. The 
        definitive filename will be built using the provided
        'filename' and adding some more information in the name.

        This method will generate a temporary filename that uses
        the current segment index in its name and is placed in 
        segment parts folder.

        This method returns the final filename created.
        """
        if not filename:
            raise Exception('No "filename" provided.')

        temp_filename = get_temp_filename(filename)

        return f'{DEFAULT_SEGMENT_PARTS_FOLDER}/segment_{self.index}_{temp_filename}'
    
    def validate(self):
        """
        Validates that the main data of the Element is correctly
        set, all required fields are provided, there are no
        unexpected shortcodes and more stuff.

        This method will raise an Exception if something is wrong.
        """
        # TODO: Move this to a external validator (?)
        # If 'text' have shortcodes
        self.shortcode_parser.parse(self.text)

        # If 'narration_text' have invalid shortcodes
        self.shortcode_parser.parse(self.narration_text)

        # If element has not necessary parameters
        self.rules_checker.check_this_need_rules(self.data)

    def create_narration(self):
        """
        Creates the audio narration (if needed) by generating an AI audio
        narration with provided 'voice' and 'audio_narration_text'
        parameters or by using the 'audio_narration_filename'.

        This method will set the 'audio_filename' to be able to build the
        audio clip in a near future.
        """
        if self.audio_narration_filename:
            segment_part_filename = self.create_segment_file(f'narration.{get_file_extension(self.audio_narration_filename)}')
            copy_file(self.audio_narration_filename, segment_part_filename)
            print_completed('Original voice narration file copied to segment parts folder')
            self.audio_narration_filename = segment_part_filename
            self.audio_filename = segment_part_filename
        else:
            segment_part_filename = self.create_segment_file('narration.wav')
            # TODO: Voice parameter need to change
            self.audio_filename = self.builder.build_narration(self.narration_text_sanitized_without_shortcodes, output_filename = segment_part_filename)
            print_completed('Voice narration created successfully')

    def create_transcription(self):
        """
        Creates the transcription of the generated audio narration
        that would be stored in 'self.audio_filename'.
        
        This method returns a words array containing, for each word,
        a 'text', 'start' and 'end' field to be able to use the 
        transcription timestamps.
        """
        transcription = get_transcription_with_timestamps(self.audio_filename, initial_prompt = self.narration_text_sanitized_without_shortcodes)

        # We join all words together as we don't need segments, just words with time
        words = []
        for transcription_segment in transcription['segments']:
            words += transcription_segment['words']
        self.transcription = words


# TODO: Remove this in a near future

# I don't remove this directly because some of the structure
# and logic applied in these classes are interesting but due
# to some issues with the inheritance I decided to simplify
# the code, but I need to refactor and improve it


from youtube_autonomous.shortcodes.enums import ShortcodeTagType
from youtube_autonomous.segments.enums import ShortcodeStart, ShortcodeDuration, ShortcodeField, ShortcodeType, ShortcodeMode
from youtube_autonomous.experimental.enhancement.enhancement_element import EnhancementElement
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.elements.validator.element_validator import ElementValidator
from youtube_autonomous.config import Configuration
from youtube_autonomous.experimental.shortcodes.consts import FILE_DURATION
from typing import Union


LOWER_INDEX = -1
UPPER_INDEX = 99999

class Shortcode:
    """
    Class that represent a shortcode detected in a text, containing 
    its attributes and values and also the content if block-scoped.
    """
    _tag: ShortcodeType
    """
    The shortcode tag that represents and identifies it. It
    determines the way it will be built and applied.
    """
    _type: ShortcodeTagType
    _start: float
    _duration: float
    _keywords: str
    _filename: str
    _url: str
    _mode: ShortcodeMode
    _context: any # TODO: Do I really need this (?)
    _content: str
    _previous_start_word_index: int
    _previous_end_word_index: int

    def __init__(self, tag: ShortcodeType, type: ShortcodeTagType, attributes: dict, context, content: str):
        """
        The shortcode has a 'type' and could include some 'attributes' that
        are the parameters inside the brackets, that could also be simple or
        include an associated value. If it is a block-scoped shortcode, it
        will have some 'content' inside of it.
        """
        self.tag = tag
        self.type = type

        self.config = Configuration.get_configuration_by_type(self.tag).config_as_shortcode
        
        self.start = attributes.get(ShortcodeField.START.value, None)
        self.duration = attributes.get(ShortcodeField.DURATION.value, None)
        self.keywords = attributes.get(ShortcodeField.KEYWORDS.value, '')
        self.filename = attributes.get(ShortcodeField.FILENAME.value, '')
        self.url = attributes.get(ShortcodeField.URL.value, '')
        self.mode = attributes.get(ShortcodeField.MODE.value, None)

        # TODO: Should I treat these shortcodes as Enhancement so I need
        # to make the same checks (more specific that these ones below) (?)

        if not self.keywords and not self.filename and not self.url:
            raise Exception('No "keywords" nor "filename" nor "url" sources available.')

        # TODO: Do we actually need the context (?)
        self.context = context
        # TODO: Do we actually need the content (?)
        self.content = content

        self.previous_start_word_index = None
        self.previous_end_word_index = None

        # TODO: Maybe we need to create an abstract Shortcode class to inherit
        # in specific shortcode classes

    @property
    def tag(self):
        """
        The shortcode tag that represents and identifies it. It
        determines the way it will be built and applied.
        """
        return self._tag
    
    @tag.setter
    def tag(self, tag: Union[ShortcodeType, str]):
        self._tag = ShortcodeType.to_enum(tag)

    @property
    def type(self):
        """
        The type of the shortcode, that identifies its structure and
        parameters it must have.
        """
        return self._type
    
    @type.setter
    def type(self, type: Union[ShortcodeTagType, str]):
        self._type = ShortcodeTagType.to_enum(type)
        
    @property
    def start(self):
        """
        The time moment of the current segment in which this element is
        expected to be applied.
        """
        return self.start
    
    @start.setter
    def start(self, start: Union[ShortcodeStart, int, float, None]):
        if start is None:
            if self.type == ShortcodeTagType.BLOCK:
                start = ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD
            else:
                start = ShortcodeStart.BETWEEN_WORDS

        self._start = ElementValidator.validate_shortcode_start_field(start)

    @property
    def duration(self):
        """
        The duration of the shortcode, that it is calculated according to the
        field or to its content.
        """
        return self._duration
    
    @duration.setter
    def duration(self, duration: Union[ShortcodeDuration, int, float, None]):
        if duration is None:
            if self.type == ShortcodeTagType.BLOCK:
                duration = ShortcodeDuration.SHORTCODE_CONTENT
            else:
                duration = ShortcodeDuration.FILE_DURATION

        # TODO: I'm setting "file_duration" but checking if it is 99998 value
        # so it is failing. Should I put the numeric value here (?)
        self._duration = ElementValidator.validate_shortcode_duration_field(duration.name)

    @property
    def mode(self):
        return self._mode
    
    @mode.setter
    def mode(self, mode: Union[ShortcodeMode, str]):
        ShortcodeMode.to_enum(mode)

        if mode not in self.config.modes:
            raise Exception(f'The provided "{self.mode.value}" is not a valid mode for this shortcode type "{self.type.value}" and tag "{self.tag.value}".')

        self._mode = mode
    
    @property
    def context(self):
        """
        The context of the shortcode.

        TODO: Do I really need this (?)
        """
        return self._context
    
    @context.setter
    def context(self, context: any):
        # TODO: Do I need to check something (?)

        self._context = context

    @property
    def content(self):
        """
        The text that is between the shortcode open and end tag
        and can include shortcodes. This parameter makes sense if
        the shortcode is a block-scoped shortcode.
        """
        return self._content
    
    @content.setter
    def content(self, content: str):
        # TODO: Do I need to check something (?)
        if content is not None and not isinstance(content, str):
            raise Exception('The "content" parameter provided is not a string.')
        
        if content is None:
            content = ''

        self._content = content

    @property
    def previous_start_word_index(self):
        """
        The index, obtained from the whole text of the segment, of the
        word that is inmediately before the shortcode start tag.
        """
        return self._previous_start_word_index
    
    @previous_start_word_index.setter
    def previous_start_word_index(self, previous_start_word_index: int):
        if not previous_start_word_index and previous_start_word_index != 0:
            raise Exception('No "previous_start_word_index" provided.')
        
        if not isinstance(previous_start_word_index, int):
            raise Exception('The "previous_start_word_index" parameter provided is not an int number.')

        if previous_start_word_index < LOWER_INDEX or previous_start_word_index > UPPER_INDEX:
            raise Exception(f'No valid "previous_start_word_index" parameter provided. Must be between {LOWER_INDEX} and {UPPER_INDEX}.')
        
        self._previous_start_word_index = previous_start_word_index

    @property
    def previous_end_word_index(self):
        """
        The index, obtained from the whole text of the segment, of the
        word that is inmediately before the shortcode end tag.
        """
        return self._previous_end_word_index
    
    @previous_end_word_index.setter
    def previous_end_word_index(self, previous_end_word_index: int):
        if not previous_end_word_index and previous_end_word_index != 0:
            raise Exception('No "previous_end_word_index" provided.')
        
        if not isinstance(previous_end_word_index, int):
            raise Exception('The "previous_end_word_index" parameter provided is not an int number.')
        
        if previous_end_word_index < LOWER_INDEX or previous_end_word_index > UPPER_INDEX:
            raise Exception(f'No valid "previous_end_word_index" parameter provided. Must be between {LOWER_INDEX} and {UPPER_INDEX}.')
        
        self._previous_end_word_index = previous_end_word_index

    def __calculate_start_and_duration(self, transcription):
        """
        Processes this shortcode 'start' and 'duration' fields by using
        the 'transcription' if needed (if 'start' and 'duration' fields
        are not numbers manually set by the user in the shortcode when
        written).

        This will consider the current 'start' and 'duration' strategy
        and apply them to the words related to the shortcode to obtain
        the real 'start' and 'duration' number values.
        """
        if isinstance(self.start, ShortcodeStart):
            # TODO: What if single shortcode with no next word (?)
            if self.type == ShortcodeTagType.SIMPLE:
                if self.start == ShortcodeStart.BETWEEN_WORDS:
                    self.start = (transcription[self._previous_start_word_index]['end'] + transcription[self._previous_start_word_index + 1]['start']) / 2
                # TODO: What about block-scoped value when simple type (?)
            else:
                if self.start == ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD:
                    self.start = transcription[self._previous_start_word_index + 1]['start']
                elif self.start == ShortcodeStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD:
                    self.start = (transcription[self._previous_start_word_index + 1]['start'] + transcription[self._previous_start_word_index + 1]['end']) / 2
                elif self.start == ShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD:
                    self.start = transcription[self._previous_start_word_index + 1]['end']
                # TODO: What about simple value when block-scoped type (?)

        if isinstance(self.duration, ShortcodeDuration):
            if self.type == ShortcodeTagType.SIMPLE:
                if self.duration == ShortcodeDuration.FILE_DURATION:
                    # This duration must be set when the file is ready, so 
                    # we use a number value out of limits to flag it
                    self.duration = FILE_DURATION
            else:
                if self.duration == ShortcodeDuration.SHORTCODE_CONTENT:
                    self.duration = transcription[self.previous_end_word_index]['end'] - transcription[self.previous_start_word_index + 1]['start']

    def to_enhancement_element(self, transcription):
        """
        Turns the current shortcode to an EnhancementElement by using
        the provided 'transcription' and using its words to set the
        actual 'start' and 'duration' fields according the narration.

        The provided 'transcription' could be not needed if the segment
        is not narrated and 'start' and 'duration' fields are manually
        set by the user in the shortcode.
        """
        if self.type == ShortcodeTagType.SIMPLE and self.previous_start_word_index is None:
            raise Exception(f'Found {ShortcodeTagType.SIMPLE.value} shortcode without "previous_start_word_index".')
        
        if self.type == ShortcodeTagType.BLOCK and (self.previous_start_word_index is None or self.previous_end_word_index is None):
            raise Exception(f'Found {ShortcodeTagType.BLOCK.value} shortcode without "previous_start_word_index" or "previous_end_word_index".')
        
        self.__calculate_start_and_duration(transcription)

        # TODO: Build it
        enhancement_element = EnhancementElement.get_class_from_type(self.tag)(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)

        # TODO: Remove this below if the code above is working
        # if self.tag == ShortcodeType.MEME:
        #     enhancement_element = MemeEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # elif self.tag == ShortcodeType.SOUND:
        #     enhancement_element = SoundEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # elif self.tag == ShortcodeType.IMAGE:
        #     enhancement_element = ImageEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # elif self.tag == ShortcodeType.STICKER:
        #     enhancement_element = StickerEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # elif self.tag == ShortcodeType.GREEN_SCREEN:
        #     enhancement_element = GreenscreenEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # elif self.tag == ShortcodeType.EFFECT:
        #     enhancement_element = EffectEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
        # else:
        #     raise Exception(f'No valid shortcode "{self.tag}" type provided.')
        #     # TODO: Implement the other EnhancementElements
        #     enhancement_element = EnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)

        return enhancement_element


"""
This file contains some constants that are useful for validation.
"""
from youtube_autonomous.segments.enums import SegmentType


# Valid segment types for validation
NARRATION_SEGMENT_TYPES = [SegmentType.CUSTOM_STOCK.value, SegmentType.STOCK.value, SegmentType.AI_IMAGE.value, SegmentType.IMAGE.value, SegmentType.YOUTUBE_VIDEO.value, SegmentType.TEXT.value]
URL_SEGMENT_TYPES = [SegmentType.IMAGE.value, SegmentType.YOUTUBE_VIDEO.value]
CUSTOM_SEGMENT_TYPES = [SegmentType.MEME.value, SegmentType.PREMADE.value]
VALID_SEGMENT_TYPES = NARRATION_SEGMENT_TYPES + URL_SEGMENT_TYPES + CUSTOM_SEGMENT_TYPES

# Arrays for validation
NEED_KEYWORDS_SEGMENT_TYPES = [SegmentType.CUSTOM_STOCK.value, SegmentType.STOCK.value, SegmentType.AI_IMAGE.value, SegmentType.MEME.value]
NEED_TEXT_SEGMENT_TYPES = [SegmentType.PREMADE.value, SegmentType.TEXT.value]


# TODO: Remove this in a near future

from youtube_autonomous.segments.enums import EnhancementType, EnhancementMode
from youtube_autonomous.experimental.segments.enhancements.validation.segment_enhancement_validator import SegmentEnhancementValidator
from youtube_autonomous.experimental.shortcodes.consts import FILE_DURATION, SEGMENT_DURATION
from moviepy.editor import AudioFileClip, CompositeAudioClip, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, concatenate_videoclips, concatenate_audioclips
from typing import Union


class EnhancementElement:
    """
    Class that represents an element that will be used to enhance a 
    segment video by applying it. It could be an effect, a sound, a
    sticker, etc.

    This class can be built directly from a json file that is read
    as a dict, or by a previously detected shortcode that is converted
    into this.
    """
    @classmethod
    def get_class_from_type(cls, type: Union[EnhancementType, str]):
        """
        This method returns the EnhancementElement class that is
        equivalent of the provided 'type' EnhancementType.
        """
        SegmentEnhancementValidator.validate_type(type)

        if isinstance(type, str):            
            type = EnhancementType(type)

        from youtube_autonomous.experimental.segments.enhancements.objects.meme_enhancement_element import MemeEnhancementElement
        from youtube_autonomous.experimental.segments.enhancements.objects.sound_enhancement_element import SoundEnhancementElement
        from youtube_autonomous.experimental.segments.enhancements.objects.image_enhancement_element import ImageEnhancementElement
        from youtube_autonomous.experimental.segments.enhancements.objects.sticker_enhancement_element import StickerEnhancementElement
        from youtube_autonomous.experimental.segments.enhancements.objects.greenscreen_enhancement_element import GreenscreenEnhancementElement
        from youtube_autonomous.experimental.segments.enhancements.objects.effect_enhancement_element import EffectEnhancementElement

        if type == EnhancementType.MEME:
            return MemeEnhancementElement
        if type == EnhancementType.SOUND:
            return SoundEnhancementElement
        if type == EnhancementType.IMAGE:
            return ImageEnhancementElement
        if type == EnhancementType.STICKER:
            return StickerEnhancementElement
        if type == EnhancementType.GREENSCREEN:
            return GreenscreenEnhancementElement
        if type == EnhancementType.EFFECT:
            return EffectEnhancementElement
        
        raise Exception('The provided "type" does not have a related EnhancementElement class.')

    _type: EnhancementType
    """
    The enhancement element type that determines the way it will be
    built and applied.
    """
    _start: float
    """
    The time moment of the current segment in which this element is
    expected to be applied.
    """
    _duration: float
    """
    The duration of the element that the user wants to be applied.
    """
    _keywords: str = ''
    """
    The keywords that will be used to look for the resource in the 
    platform that the type of element needs, if provided.
    """
    _url: str = ''
    """
    The url from which the resource will be obtained, if set, to
    use as source of this enhancement element.
    """
    _filename: str = ''
    """
    The filename that will be used as source of this enhancement 
    element if provided.
    """
    _mode: EnhancementMode = EnhancementMode.INLINE
    """
    The mode in which the enhancement element will be applied, that could be inline
    or as overview.
    """

    def __init__(self, type: EnhancementType, start: float, duration: float, keywords: str, url: str, filename: str, mode: EnhancementMode):
        self.type = type
        self.start = start
        self.duration = duration
        self.keywords = keywords
        self.url = url
        self.filename = filename
        self.mode = mode

    @property
    def valid_modes(self):
        """
        The EnhancementMode enums accepted for this type of
        EnhancementElement.
        """
        return EnhancementMode.get_all()
    
    @valid_modes.setter
    def valid_modes(self, value):
        raise Exception('You cannot modify the "valid_modes" parameter.')

    @property
    def type(self):
        """
        The enhancement element type that determines the way it will be
        built and applied.
        """
        return self._type
    
    @type.setter
    def type(self, type: Union[EnhancementType, str]):
        SegmentEnhancementValidator.validate_type(type)
            
        if isinstance(type, str):
            type = EnhancementType(type)

        self._type = type

    @property
    def start(self):
        """
        The time moment of the current segment in which this element is
        expected to be applied.
        """
        return self.start
    
    @start.setter
    def start(self, start: float):
        SegmentEnhancementValidator.validate_start(start)
        
        self._start = start

    @property
    def duration(self):
        """
        The duration of the element that the user wants to be applied.
        """
        return self._duration
    
    @duration.setter
    def duration(self, duration: float):
        SegmentEnhancementValidator.validate_duration(duration)
        
        self._duration = duration

    @property
    def keywords(self):
        """
        The keywords that will be used to look for the resource in
        the platform that the type of element needs, if provided.
        """
        return self._keywords
    
    @keywords.setter
    def keywords(self, keywords: Union[str, None]):
        if keywords is not None and keywords != '':
            if not keywords:
                raise Exception('No "keywords" provided.')
            
            if not isinstance(keywords, str):
                raise Exception(f'The "keywords" parameter provided {str(keywords)} is not a string.')
        else:
            keywords = ''
        
        self._keywords = keywords

    @property
    def url(self):
        """
        The url from which the resource will be obtained, if set, to
        use as source of this enhancement element.
        """
        return self._url
    
    @url.setter
    def url(self, url: Union[str, None]):
        if url is not None and url != '':
            if not url:
                raise Exception('No "url" provided.')
            
            if not isinstance(url, str):
                raise Exception(f'The "url" parameter provided {str(url)} is not a string.')
        else:
            url = ''
        
        self._url = url

    @property
    def filename(self):
        """
        The filename that will be used as source of this enhancement 
        element if provided.
        """
        return self._filename
    
    @filename.setter
    def filename(self, filename: Union[str, None]):
        if filename is not None and filename != '':
            if not filename:
                raise Exception('No "filename" provided.')
            
            if not isinstance(filename, str):
                raise Exception(f'The "filename" parameter provided {str(filename)} is not a string.')
        else:
            filename = ''
        
        self._filename = filename

    @property
    def mode(self):
        """
        The mode in which the enhancement element will be applied, that could be inline
        or as overview.
        """
        return self._mode
    
    @mode.setter
    def mode(self, mode: Union[EnhancementMode, str]):
        SegmentEnhancementValidator.validate_mode(mode, self.valid_modes)
        
        if isinstance(mode, str):
            mode = EnhancementMode(mode)

        self._mode = mode


    def recalculate_duration(self, enhancement_clip: Union[AudioFileClip, CompositeAudioClip, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        This method recalculates the 'self.duration' field according to
        its value, the 'self.mode' parameter value we have, and also the
        'enhancement_clip' and 'video' parameters provided.
        """
        if self.duration == FILE_DURATION:
            self.duration = enhancement_clip.duration

        if self.duration == SEGMENT_DURATION:
            self.duration = video.duration

        if self.mode in [EnhancementMode.INLINE, EnhancementMode.OVERLAY]:
            if self.duration > enhancement_clip.duration:
                self.duration = enhancement_clip.duration

        if self.mode == [EnhancementMode.OVERLAY, EnhancementMode.REPLACE]:
            end = self.start + self.duration
            if end > video.duration:
                self.duration = video.duration - self.start

    def apply(self, enhancement_clip: Union[AudioFileClip, CompositeAudioClip, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        This method adjusts the 'enhancement_clip' duration and make the
        necessary cuts to fit in the base provided 'video' according to
        its parameters. It will consider the mode to cut the 'video' and
        insert the 'enhancement_clip' in some specific position, or will
        put it as an overlay video if requested.

        This method returns the provided 'video' with the 
        'enhancement_clip' applied on it.
        """
        if isinstance(enhancement_clip, (AudioFileClip, CompositeAudioClip)):
            audio = video.audio
            if self.mode == EnhancementMode.INLINE:
                if self.start == 0:
                    audio = concatenate_audioclips([
                        enhancement_clip.set_duration(self.duration),
                        audio
                    ])
                elif self.start == video.duration:
                    audio = concatenate_audioclips([
                        audio,
                        enhancement_clip.set_duration(self.duration)
                    ])
                else:
                    audio = concatenate_audioclips([
                        audio.subclip(0, self.start),
                        enhancement_clip.set_duration(self.duration),
                        audio.subclip(self.start, video.duration)
                    ])
            elif self.mode in [EnhancementMode.OVERLAY, EnhancementMode.REPLACE]:
                audio = CompositeAudioClip([
                    audio,
                    enhancement_clip.set_start(self.start).set_duration(video.duration)
                ])
            # TODO: Implement self.mode == EnhancementMode.REPLACE (?)

            video = video.set_audio(audio)
        elif isinstance(enhancement_clip, (VideoFileClip, CompositeVideoClip, ImageClip, ColorClip)):
            if self.mode == EnhancementMode.INLINE:
                if self.start == 0:
                    video = concatenate_videoclips([
                        enhancement_clip.set_duration(self.duration),
                        video
                    ])
                elif self.start == video.duration:
                    video = concatenate_videoclips([
                        video,
                        enhancement_clip.set_duration(self.duration)
                    ])
                else:
                    video = concatenate_videoclips([
                        video.subclip(0, self.start),
                        enhancement_clip.set_duration(self.duration),
                        video.subclip(self.start, video.duration)
                    ])
            elif self.mode == EnhancementMode.OVERLAY:
                video = CompositeVideoClip([
                    video,
                    enhancement_clip.set_start(self.start).set_duration(video.duration)
                ])
            elif self.mode == EnhancementMode.REPLACE:
                end = self.start + self.duration
                if self.start == 0:
                    if end == video.duration:
                        video = enhancement_clip
                    else:
                        video = concatenate_videoclips([
                            enhancement_clip,
                            video.subclip(end, video.duration)
                        ])
                elif end == video.duration:
                    video = concatenate_videoclips([
                        video.subclip(0, self.start),
                        enhancement_clip
                    ])
                else:
                    video = concatenate_videoclips([
                        video.subclip(0, self.start),
                        enhancement_clip,
                        video.subclip(end, video.duration)
                    ])

        return video
        


from setuptools import setup, find_packages

VERSION = '0.0.51'
DESCRIPTION = 'Youtube Autonomous main software.'
LONG_DESCRIPTION = 'This is the main Youtube Autonomous project handler, containing the whole functionality.'

setup(
        name = "youtube-autonomous", 
        version = VERSION,
        author = "Daniel Alcalá",
        author_email = "<danielalcalavalera@gmail.com>",
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        packages = find_packages(),
        install_requires = [
            'yta-multimedia',
            'yta-general-utils',
            'youtubeenhanced',
            'yta-stock-downloader'
        ],
        
        keywords = [
            'youtube autonomous main software'
        ],
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_multimedia.video.generation.manim.classes.text.text_word_by_word_manim_animation import TextWordByWordManimAnimation
from yta_multimedia.video.generation.manim.classes.text.text_triplets_manim_animation import TextTripletsManimAnimation
from yta_general_utils.programming.parameter_obtainer import ParameterObtainer
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.editor import VideoFileClip
from typing import Union


# I don't know exactly how to manage this because I have
# classes and then the name that actually I make to be
# corresponding to that class
class TextPremade(Enum):
    TEXT_TRIPLETS = 'text_triplets'

class TextElementBuilder(ElementBuilder):
    @classmethod
    def build_custom_from_text_class_name(cls, text_class_name: str, **parameters):
        ElementParameterValidator.validate_text_class_name(text_class_name)

        text_class = None
        if text_class == 'text_word_by_word':
            text_class = TextWordByWordManimAnimation
        else:
            raise Exception(f'The provided "text_class" parameter {text_class} is not a valid text class name.')
        
        return cls.build_custom(text_class, **parameters)

    @classmethod
    def build_from_enhancement(cls, enhancement: dict):
        text_premade_name = enhancement.keywords

        ElementParameterValidator.validate_premade_name(text_premade_name)

        text_premade_class = cls.text_premade_name_to_class(text_premade_name)
        parameters_to_ignore = ['self', 'cls', 'args', 'kwargs', 'output_filename', 'duration']
        parameters = super().extract_extra_params(enhancement, text_premade_class.generate, parameters_to_ignore)

        # For this specific case, we use the 'duration' field as the
        # duration extra_param if needed, because it could be using
        # a narration so the actual 'duration' must be that
        parameters['duration'] = enhancement.duration

        actual_parameters = ParameterObtainer.get_parameters_from_method(text_premade_class.generate)

        # I need to ensure that parameters has only 'mandatory' or 'optional'
        # actual parameters of the method
        parameters = {key: value for key, value in parameters.items() if key in actual_parameters['mandatory'] or key in actual_parameters['optional']}

        return cls.build_custom_from_text_premade_name(text_premade_name, **parameters)

    @classmethod
    def build_from_segment(cls, segment: dict):
        text_premade_name = segment.keywords

        ElementParameterValidator.validate_premade_name(text_premade_name)

        text_premade_class = cls.text_premade_name_to_class(text_premade_name)
        parameters_to_ignore = ['self', 'cls', 'args', 'kwargs', 'output_filename', 'duration']
        parameters = super().extract_extra_params(segment, text_premade_class.generate, parameters_to_ignore)

        # For this specific case, we use the 'duration' field as the
        # duration extra_param if needed, because it could be using
        # a narration so the actual 'duration' must be that
        parameters['duration'] = segment.duration

        actual_parameters = ParameterObtainer.get_parameters_from_method(text_premade_class.generate)

        # I need to ensure that parameters has only 'mandatory' or 'optional'
        # actual parameters of the method
        parameters = {key: value for key, value in parameters.items() if key in actual_parameters['mandatory'] or key in actual_parameters['optional']}

        return cls.build_custom_from_text_premade_name(text_premade_name, **parameters)
    
    @classmethod
    def text_premade_name_to_class(cls, text_premade_name: str):
        """
        Returns the corresponding text premade class according to
        the provided 'text_premade_name'. If no text premade class
        found, the return will be None.
        """
        text_premade_class = None

        if text_premade_name == TextPremade.TEXT_TRIPLETS.value:
            text_premade_class = TextTripletsManimAnimation
        else:
            raise Exception(f'The provided text premade name "{text_premade_name}" is not valid. The valid ones are: {TextPremade.get_all_values_as_str()}')
        
        return text_premade_class

    @classmethod
    def build_custom_from_text_premade_name(cls, text_premade_name, **parameters):
        """
        This method instantiates the 'text_animation_class' Manim
        text animation class and uses the provided 'parameters' to
        build the text animation. The provided 'parameters' must 
        fit the ones requested by the provided class 'generate'
        method.
        """
        ElementParameterValidator.validate_premade_name(text_premade_name)

        text_premade_class = cls.text_premade_name_to_class(text_premade_name)

        # We generate the animation to return it
        filename = text_premade_class().generate(**parameters)

        return VideoFileClip(filename)

    @classmethod
    def build(cls, text: str, duration: Union[int, float]):
        """
        Basic example to test that the building process and
        the class are working correctly.

        TODO: Remove this in the future when 'custom' is 
        working perfectly.
        """
        ElementParameterValidator.validate_text(text)
        ElementParameterValidator.validate_duration(duration)

        filename = TextWordByWordManimAnimation().generate(text, duration)

        video = VideoFileClip(filename)

        return video


from yta_general_utils.programming.enum import YTAEnum as Enum


class EditionManualTermMode(Enum):
    """
    This is the mode in which we will look for the terms
    of our edition manual book in the given segment text
    to find any coincidences.
    """
    EXACT = 'exact'
    """
    The term found must be exactly matching one term of
    our edition manual book.
    """
    IGNORE_CASE_AND_ACCENTS = 'ignore_case_and_accents'
    """
    The term found must match, in lower case and ignoring
    the accents, one term of our edition manual book.
    """
    @classmethod
    def get_default(cls):
        """
        Returns the item that acts as the one by default.
        """
        return cls.EXACT

class EditionManualTermContext(Enum):
    """
    This is the context we will be able to apply to our
    edition terms to be applied only on those segments
    related to that context.
    """
    ANY = 'any'
    """
    The term will be applied always, in any context.
    """
    @classmethod
    def get_default(cls):
        """
        Returns the item that acts as the one by default.
        """
        return cls.ANY


from youtube_autonomous.segments.enums import Type, SegmentType, EnhancementType, ShortcodeType, SegmentMode, EnhancementMode, ShortcodeMode, SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_general_utils.programming.decorators import classproperty
from abc import ABCMeta
from typing import Union


def get_configuration(type: Union[str, Type]):
    type = Type.to_enum(type)

    if type == Type.STOCK:
        return StockConfiguration
    
class SegmentConfiguration:
    """
    The configuration of an element as a Segment.
    """
    type: SegmentType
    modes: list[SegmentMode]
    default_mode: SegmentMode
    string_durations: list[SegmentStringDuration]
    """
    A list containing the StringDuration values that we accept
    in this segment configuration.
    """

    def __init__(self, type: SegmentType, string_durations = None):
        self.type = type,
        self.modes = None, # By now we don't accept modes
        self.default_mode = None # By now we don't accept modes
        self.string_durations = None 

        if string_durations is not None:
            self.string_durations = string_durations

    # TODO: Should we keep this below (?)
    @staticmethod
    def is_valid_type(segment_type: Union[str, SegmentType]):
        """
        Validates if the provided 'segment_type' is a valid name or 
        value of the registered and valid SegmentType enums. This
        method will return the Enum item if valid, or False if not.
        """
        return SegmentType.is_valid(segment_type)

    @staticmethod
    def is_valid_duration(duration: Union[float, int, str, SegmentStringDuration]):
        if isinstance(duration, str):
            return SegmentStringDuration.is_valid(duration)
        ElementParameterValidator.validate_numeric_positive_mandatory_parameter('duration', duration)
    # TODO: Should we keep this above (?)

class EnhancementConfiguration:
    """
    The configuration of an element as an Enhancement.
    """
    type: EnhancementType
    modes: list[EnhancementMode]
    default_mode: EnhancementMode
    string_durations: list[EnhancementStringDuration]

    def __init__(self, type: EnhancementType, modes: list[EnhancementMode], default_mode: EnhancementMode, string_durations: list[EnhancementStringDuration]):
        self.type = type
        self.modes = modes
        self.default_mode = default_mode
        self.string_durations = string_durations

class ShortcodeConfiguration:
    """
    The configuration of an element as a Shortcode.
    """
    type: ShortcodeType
    modes: list[ShortcodeMode]
    default_mode: ShortcodeMode
    string_durations: list[ShortcodeStringDuration]

    def __init__(self, type: EnhancementType, modes: list[ShortcodeMode], default_mode: ShortcodeMode, string_durations: list[ShortcodeStringDuration]):
        self.type = type
        self.modes = modes
        self.default_mode = default_mode
        self.string_durations = string_durations

class Configuration(metaclass = ABCMeta):
    _type: Type

    _can_have_narration: bool
    _need_narration: bool
    _can_have_specific_duration: bool
    _need_specific_duration: bool
    _can_have_text: bool
    _need_text: bool
    _can_have_filename: bool
    _can_have_url: bool
    _need_filename_or_url: bool
    _can_have_keywords: bool
    _need_keywords: bool
    _can_have_more_parameters: bool

    _config_as_segment: SegmentConfiguration
    _config_as_enhancement: EnhancementConfiguration
    _config_as_shortcode: ShortcodeConfiguration

    @classproperty
    def type(cls):
        return cls._type

    @classproperty
    def can_have_narration(cls):
        return cls._can_have_narration
    
    @classproperty
    def need_narration(cls):
        return cls._need_narration
    
    @classproperty
    def can_have_specific_duration(cls):
        return cls._can_have_specific_duration
    
    @classproperty
    def need_specific_duration(cls):
        return cls._need_specific_duration
    
    @classproperty
    def can_have_text(cls):
        return cls._can_have_text
    
    @classproperty
    def need_text(cls):
        return cls._need_text
    
    @classproperty
    def can_have_filename(cls):
        return cls._can_have_filename
    
    @classproperty
    def can_have_url(cls):
        return cls._can_have_url
    
    @classproperty
    def need_filename_or_url(cls):
        return cls._need_filename_or_url

    @classproperty
    def can_have_keywords(cls):
        return cls._can_have_keywords
    
    @classproperty
    def need_keywords(cls):
        return cls._need_keywords
    
    @classproperty
    def can_have_more_parameters(cls):
        return cls._can_have_more_parameters
    
    @classproperty
    def config_as_segment(cls):
        return cls._config_as_segment
    
    @classproperty
    def config_as_enhancement(cls):
        return cls._config_as_enhancement
    
    @classproperty
    def config_as_shortcode(cls):
        return cls._config_as_shortcode
    
    @staticmethod
    def get_configuration_by_type(type: Union[str, Type]):
        """
        Returns the configuration object that corresponds to
        the provided 'type' that need to be a valid type string
        or object.
        """
        if isinstance(type, (SegmentType, EnhancementType, ShortcodeType)):
            type = Type.to_enum(type.value)
        else:
            type = Type.to_enum(type)

        if type == Type.STOCK:
            return StockConfiguration
        elif type == Type.CUSTOM_STOCK:
            return CustomStockConfiguration
        elif type == Type.AI_IMAGE:
            return AIImageConfiguration
        elif type == Type.IMAGE:
            return ImageConfiguration
        elif type == Type.AI_VIDEO:
            return AIVideoConfiguration
        elif type == Type.VIDEO:
            return VideoConfiguration
        elif type == Type.SOUND:
            return SoundConfiguration
        elif type == Type.YOUTUBE_VIDEO:
            return YoutubeVideoConfiguration
        elif type == Type.TEXT:
            return TextConfiguration
        elif type == Type.MEME:
            return MemeConfiguration
        elif type == Type.EFFECT:
            return EffectConfiguration
        elif type == Type.PREMADE:
            return PremadeConfiguration
        elif type == Type.GREENSCREEN:
            return GreenscreenConfiguration

class StockConfiguration(Configuration):
    """
    The 'Stock' element configuration, that defines if it can be used
    as a segment or as a enhancement (or both) and all the things it
    needs to work as expected.
    """
    _type = Type.STOCK
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = False
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.STOCK,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.STOCK,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.STOCK,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class CustomStockConfiguration(Configuration):
    """
    The 'CustomStock' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.CUSTOM_STOCK
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = False
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.CUSTOM_STOCK,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.CUSTOM_STOCK,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.CUSTOM_STOCK,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class AIImageConfiguration(Configuration):
    """
    The 'AIImage' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.AI_IMAGE
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.AI_IMAGE,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.AI_IMAGE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.AI_IMAGE,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class ImageConfiguration(Configuration):
    """
    The 'Image' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.IMAGE
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _need_keywords: bool = False
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.IMAGE,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.IMAGE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.IMAGE,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')

class AIVideoConfiguration(Configuration):
    """
    The 'AIVideo' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.AI_VIDEO
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.AI_VIDEO,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.AI_VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.AI_VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class VideoConfiguration(Configuration):
    """
    The 'Video' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.VIDEO
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _need_keywords: bool = False
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.VIDEO,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class SoundConfiguration(Configuration):
    """
    The 'Sound' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.SOUND
    
    _can_have_narration = False
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = False
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _need_filename_or_url: bool = True
    _can_have_keywords: bool = True
    _need_keywords: bool = False
    _can_have_more_parameters: bool = False

    _config_as_segment: None
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.SOUND,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION, EnhancementStringDuration.FILE_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.SOUND,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT, ShortcodeStringDuration.FILE_DURATION]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')

class YoutubeVideoConfiguration(Configuration):
    """
    The 'YoutubeVideo' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.YOUTUBE_VIDEO
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = True
    _need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _need_keywords: bool = False
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.YOUTUBE_VIDEO,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.YOUTUBE_VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.YOUTUBE_VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')

class TextConfiguration(Configuration):
    """
    The 'Text' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.TEXT
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = True
    _need_text: bool = True
    _can_have_filename: bool = False
    _can_have_url: bool = True
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = True

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.TEXT,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.TEXT,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.TEXT,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class MemeConfiguration(Configuration):
    """
    The 'Meme' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.MEME
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = False
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = False

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.MEME,
        string_durations = [EnhancementStringDuration.FILE_DURATION]
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.MEME,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.INLINE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION, EnhancementStringDuration.FILE_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.MEME,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.INLINE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT, ShortcodeStringDuration.FILE_DURATION]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class EffectConfiguration(Configuration):
    """
    The 'Effect' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.EFFECT
    
    _can_have_narration = False
    _need_narration: bool = False
    _can_have_specific_duration: bool = False
    _need_specific_duration: bool = False
    _can_have_text: bool = False
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = True

    _config_as_segment: None
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.EFFECT,
        modes = [EnhancementMode.REPLACE],
        default_mode = EnhancementMode.REPLACE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.EFFECT,
        modes = [ShortcodeMode.REPLACE],
        default_mode = ShortcodeMode.REPLACE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class PremadeConfiguration(Configuration):
    """
    The 'Premade' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.PREMADE
    
    _can_have_narration = True
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = True
    _can_have_text: bool = False  # dynamic parameters will be applied
    _need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _need_keywords: bool = True
    _can_have_more_parameters: bool = True

    _config_as_segment: SegmentConfiguration = SegmentConfiguration(
        type = SegmentType.PREMADE,
    )
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.PREMADE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.INLINE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.PREMADE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.INLINE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
    
class GreenscreenConfiguration(Configuration):
    """
    The 'Greenscreen' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = Type.GREENSCREEN
    
    _can_have_narration = False
    _need_narration: bool = False
    _can_have_specific_duration: bool = True
    _need_specific_duration: bool = False
    _can_have_text: bool = False  # dynamic parameters will be applied
    _need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _need_keywords: bool = False
    _can_have_more_parameters: bool = False

    _config_as_segment: None
    _config_as_enhancement: EnhancementConfiguration = EnhancementConfiguration(
        type = EnhancementType.GREENSCREEN,
        modes = [EnhancementMode.REPLACE],
        default_mode = EnhancementMode.REPLACE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION]
    )
    _config_as_shortcode: ShortcodeConfiguration = ShortcodeConfiguration(
        type = ShortcodeType.GREENSCREEN,
        modes = [EnhancementMode.REPLACE],
        default_mode = EnhancementMode.REPLACE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT]
    )

    def __init__(self):
        raise Exception('Sorry, this class is not instantiable.')
